@@ -5,6 +5,7 @@
 #include "./stack.h"
 #include "./length.h"
 #include <assert.h>
+#include <inttypes.h>
 #include <stdio.h>
 
 #define MAX_LINK_COUNT 8
@@ -12,9 +13,9 @@
 #define MAX_ITERATOR_COUNT 64
 
 #if defined _WIN32 && !defined __GNUC__
-#define inline __forceinline
+#define forceinline __forceinline
 #else
-#define inline static inline __attribute__((always_inline))
+#define forceinline static inline __attribute__((always_inline))
 #endif
 
 typedef struct StackNode StackNode;
@@ -81,9 +82,9 @@ typedef StackAction (*StackCallback)(void *, const StackIterator *);
 static void stack_node_retain(StackNode *self) {
   if (!self)
     return;
-  assert(self->ref_count > 0);
+  ts_assert(self->ref_count > 0);
   self->ref_count++;
-  assert(self->ref_count != 0);
+  ts_assert(self->ref_count != 0);
 }
 
 static void stack_node_release(
@@ -92,7 +93,7 @@ static void stack_node_release(
   SubtreePool *subtree_pool
 ) {
 recur:
-  assert(self->ref_count != 0);
+  ts_assert(self->ref_count != 0);
   self->ref_count--;
   if (self->ref_count > 0) return;
 
@@ -227,7 +228,8 @@ static void stack_node_add_link(
       // If the previous nodes are mergeable, merge them recursively.
       if (
         existing_link->node->state == link.node->state &&
-        existing_link->node->position.bytes == link.node->position.bytes
+        existing_link->node->position.bytes == link.node->position.bytes &&
+        existing_link->node->error_cost == link.node->error_cost
       ) {
         for (int j = 0; j < link.node->link_count; j++) {
           stack_node_add_link(existing_link->node, link.node->links[j], subtree_pool);
@@ -458,6 +460,17 @@ uint32_t ts_stack_version_count(const Stack *self) {
   return self->heads.size;
 }
 
+uint32_t ts_stack_halted_version_count(Stack *self) {
+  uint32_t count = 0;
+  for (uint32_t i = 0; i < self->heads.size; i++) {
+    StackHead *head = array_get(&self->heads, i);
+    if (head->status == StackStatusHalted) {
+      count++;
+    }
+  }
+  return count;
+}
+
 TSStateId ts_stack_state(const Stack *self, StackVersion version) {
   return array_get(&self->heads, version)->node->state;
 }
@@ -509,7 +522,7 @@ void ts_stack_push(
   head->node = new_node;
 }
 
-inline StackAction pop_count_callback(void *payload, const StackIterator *iterator) {
+forceinline StackAction pop_count_callback(void *payload, const StackIterator *iterator) {
   unsigned *goal_subtree_count = payload;
   if (iterator->subtree_count == *goal_subtree_count) {
     return StackActionPop | StackActionStop;
@@ -522,7 +535,8 @@ StackSliceArray ts_stack_pop_count(Stack *self, StackVersion version, uint32_t c
   return stack__iter(self, version, pop_count_callback, &count, (int)count);
 }
 
-inline StackAction pop_pending_callback(void *payload, const StackIterator *iterator) {
+
+forceinline StackAction pop_pending_callback(void *payload, const StackIterator *iterator) {
   (void)payload;
   if (iterator->subtree_count >= 1) {
     if (iterator->is_pending) {
@@ -544,7 +558,7 @@ StackSliceArray ts_stack_pop_pending(Stack *self, StackVersion version) {
   return pop;
 }
 
-inline StackAction pop_error_callback(void *payload, const StackIterator *iterator) {
+forceinline StackAction pop_error_callback(void *payload, const StackIterator *iterator) {
   if (iterator->subtrees.size > 0) {
     bool *found_error = payload;
     if (!*found_error && ts_subtree_is_error(iterator->subtrees.contents[0])) {
@@ -565,7 +579,7 @@ SubtreeArray ts_stack_pop_error(Stack *self, StackVersion version) {
       bool found_error = false;
       StackSliceArray pop = stack__iter(self, version, pop_error_callback, &found_error, 1);
       if (pop.size > 0) {
-        assert(pop.size == 1);
+        ts_assert(pop.size == 1);
         ts_stack_renumber_version(self, pop.contents[0].version, version);
         return pop.contents[0].subtrees;
       }
@@ -575,7 +589,7 @@ SubtreeArray ts_stack_pop_error(Stack *self, StackVersion version) {
   return (SubtreeArray) {.size = 0};
 }
 
-inline StackAction pop_all_callback(void *payload, const StackIterator *iterator) {
+forceinline StackAction pop_all_callback(void *payload, const StackIterator *iterator) {
   (void)payload;
   return iterator->node->link_count == 0 ? StackActionPop : StackActionNone;
 }
@@ -589,7 +603,7 @@ typedef struct {
   unsigned max_depth;
 } SummarizeStackSession;
 
-inline StackAction summarize_stack_callback(void *payload, const StackIterator *iterator) {
+forceinline StackAction summarize_stack_callback(void *payload, const StackIterator *iterator) {
   SummarizeStackSession *session = payload;
   TSStateId state = iterator->node->state;
   unsigned depth = iterator->subtree_count;
@@ -661,8 +675,8 @@ void ts_stack_remove_version(Stack *self, StackVersion version) {
 
 void ts_stack_renumber_version(Stack *self, StackVersion v1, StackVersion v2) {
   if (v1 == v2) return;
-  assert(v2 < v1);
-  assert((uint32_t)v1 < self->heads.size);
+  ts_assert(v2 < v1);
+  ts_assert((uint32_t)v1 < self->heads.size);
   StackHead *source_head = &self->heads.contents[v1];
   StackHead *target_head = &self->heads.contents[v2];
   if (target_head->summary && !source_head->summary) {
@@ -681,7 +695,7 @@ void ts_stack_swap_versions(Stack *self, StackVersion v1, StackVersion v2) {
 }
 
 StackVersion ts_stack_copy_version(Stack *self, StackVersion version) {
-  assert(version < self->heads.size);
+  ts_assert(version < self->heads.size);
   array_push(&self->heads, self->heads.contents[version]);
   StackHead *head = array_back(&self->heads);
   stack_node_retain(head->node);
@@ -741,7 +755,7 @@ bool ts_stack_is_paused(const Stack *self, StackVersion version) {
 
 Subtree ts_stack_resume(Stack *self, StackVersion version) {
   StackHead *head = array_get(&self->heads, version);
-  assert(head->status == StackStatusPaused);
+  ts_assert(head->status == StackStatusPaused);
   Subtree result = head->lookahead_when_paused;
   head->status = StackStatusActive;
   head->lookahead_when_paused = NULL_SUBTREE;
@@ -866,7 +880,7 @@ bool ts_stack_print_dot_graph(Stack *self, const TSLanguage *language, FILE *f)
           fprintf(f, "\"");
           fprintf(
             f,
-            "labeltooltip=\"error_cost: %u\ndynamic_precedence: %u\"",
+            "labeltooltip=\"error_cost: %u\ndynamic_precedence: %" PRId32 "\"",
             ts_subtree_error_cost(link.subtree),
             ts_subtree_dynamic_precedence(link.subtree)
           );
@@ -894,4 +908,4 @@ bool ts_stack_print_dot_graph(Stack *self, const TSLanguage *language, FILE *f)
   return true;
 }
 
-#undef inline
+#undef forceinline
